from . import Tasks, Carol
import os
import logging
import sys
import re
import math

_carol_levels = dict(
    NOTSET="NOTSET",
    DEBUG='DEBUG',
    INFO="INFO",
    WARNING="WARN",
    WARN="WARN",
    ERROR="ERROR",
    CRITICAL="ERROR",
    FATAL="ERROR",
)


class CarolHandler(logging.StreamHandler):
    """
    Carol logger handler.

    This class can be used to log information in long tasks in Carol.

    Args:
    carol: Carol object
        Carol object.

    Usage:

    >>>from pycarol import Carol, CarolHandler
    >>>import logging
    >>>logger = logging.getLogger(__name__)
    >>>carol = CarolHandler(Carol())
    >>>carol.setLevel(logging.INFO)
    >>>logger.addHandler(carol)

    >>>logger.debug('This is a debug message') #This will not be logged in Carol. Level is set to INFO
    >>>logger.info('This is an info message')
    >>>logger.warning('This is a warning message')
    >>>logger.error('This is an error message')
    >>>logger.critical('This is a critical message')

    These methods will use the current long task id provided by Carol when running your application.
    For local environments you need to set that manually first on the beginning of your code:

    >>>import os
    >>>os.environ['LONGTASKID'] = TASK_ID

    If no TASK ID is passed it works as a Console Handler.
    """

    def __init__(self, carol=None):
        """

        """
        super().__init__(stream=sys.stdout)

        self._use_console = False
        if carol is None:
            domain = os.getenv('CAROLTENANT')
            app_name = os.getenv('CAROLAPPNAME')
            auth_token = os.getenv('CAROLAPPOAUTH')
            connector_id = os.getenv('CAROLCONNECTORID')
            if ((domain is not None)
                    and (app_name is not None)
                    and (auth_token is not None)
                    and (connector_id is not None)):
                carol = Carol()
                self._use_console = False

            else:
                self._use_console = True

        self.carol = carol
        self._task = Tasks(self.carol)
        self.task_id = os.getenv('LONGTASKID', None)
        self._task.task_id = self.task_id
        self._first_pending = True

    def _log_carol(self, record):
        msg = self.format(record)
        log_level = _carol_levels.get(record.levelname)
        if 'Pending tasks' in msg:
            self._set_progress_task_luigi(msg, log_level=log_level)

        if record.name != 'luigi-interface':
            self._task.add_log(msg, log_level=log_level)

    def emit(self, record):
        if (self.task_id is None) or (self._use_console):
            super().emit(record)
        else:
            self._log_carol(record)

    def _set_progress_task_luigi(self, msg, log_level):

        """
        Used to auto set the process bar in Carol when using Luigi.

        Args:
            msg: `str`
                message to log
            log_level:
                log level.

        Returns: None

        """

        match = re.search(r'\d+.?\d*', msg)
        wrong_value = False
        if match:
            try:
                current_count = float(match.group())
            except:
                wrong_value = True

        else:
            current_count = 100
            wrong_value = True
            self._task.add_log('Something wrong with task counter', log_level='WARN')

        if (self._first_pending) and (not wrong_value):
            self._first_pending = False
            self._total_number_of_tasks = current_count

        if wrong_value:
            try:
                current_percentage = 100 - 100 * (current_count / self._total_number_of_tasks)
                current_percentage = int(min(current_percentage, 99))
                self._task.set_progress(current_percentage)
            except:
                pass
        self._task.add_log(msg, log_level='INFO')



from .query import Query
import itertools
from .data_models import DataModel
from .staging import Staging

class Tenant:
    def __init__(self, carol):
        self.carol = carol

    def get_tenant_by_domain(self, domain):
        """
        Get tenant information.

        :param domain: `str`
            Tenant name
        :return:
            dict with the information about the tenant.
        """
        return self.carol.call_api('v2/tenants/domain/{}'.format(domain), auth=False, status_forcelist=[], retries=0)



    def check_exports(self, staging=True, data_model=True):
        """
        Find all data models or stating tables that have export active.

        :param staging: `bool`, default `True`
            Get status for exports in staging.
        :param data_model: `bool`, default `True`
            Get status for exports in data models.
        :return: `tuple`
            staging stats, data model stats
        """

        staging_results = None
        dm_results = None
        if staging:
            staging_results = Staging(self.carol)._get_staging_export_stats()

        if data_model:
            dm_results = DataModel(self.carol)._get_dm_export_stats()


        return staging_results, dm_results





import json
from collections import defaultdict


class Connectors:
    """

    This class handle all APIs related to Carol connectors

    Args:
        carol: class: pycarol.Carol
    """

    def __init__(self, carol):
        self.carol = carol

    def create(self, name, label=None, group_name="Others", overwrite=False):

        """
        Create a connector

        Args:
            name: 'str'
                Connector name
            label: 'str'
                Connector label in UI
            group_name: `str` default "Others"
                Connector group name in UI
            overwrite: `bool` default `False`
                Overwrite if already exists.
        """
        if label is None:
            label = name

        resp = self.carol.call_api('v1/connectors', data={
            'mdmName': name,
            'mdmGroupName': group_name,
            'mdmLabel': {"en-US": label}
        }, errors='ignore')
        if resp.get('mdmId') is not None:
            return resp.get('mdmId')
        if ('already exists' in resp.get('errorMessage', [])):
            if overwrite:
                self.delete_by_name(name)
                return self.create(name, label, group_name, False)
            else:
                return self.get_by_name(name)['mdmId']

        else:
            raise Exception(resp)

    def get_by_name(self, name, errors='raise'):
        """
        Get connector information using the connector name

        Args:
            name: 'str'
                Connector Name
            errors: {‘ignore’, ‘raise’}, default ‘raise’
                If ‘raise’, then invalid request will raise an exception If ‘ignore’,
                then invalid request will return the request response


        Returns: `dict`
            connector information.

        """
        resp = self.carol.call_api(f'v1/connectors/name/{name}', errors=errors)
        return resp

    def get_by_id(self, id, errors='raise'):
        """
        Get connector information using the connector name

        Args:
            id: 'str'
                Connector ID
            errors: {‘ignore’, ‘raise’}, default ‘raise’
                If ‘raise’, then invalid request will raise an exception If ‘ignore’,
                then invalid request will return the request response


        Returns: `dict`
            connector information.

        """
        resp = self.carol.call_api(f'v1/connectors/{id}', errors=errors)
        return resp

    def delete_by_name(self, name, force_deletion=True):
        mdm_id = self.get_by_name(name)['mdmId']
        self.delete_by_id(mdm_id, force_deletion)

    def delete_by_id(self, mdm_id, force_deletion=True):
        self.carol.call_api('v1/connectors/{}?forceDeletion={}'.format(mdm_id, force_deletion), method='DELETE')

    def get_all(self, offset=0, page_size=-1, sort_order='ASC', sort_by=None, include_connectors=False,
                include_mappings=False, include_consumption=False, print_status=True, save_results=False,
                filename='connectors.json'):

        params = {"offset": offset, "pageSize": str(page_size), "sortOrder": sort_order,
                  "includeMappings": include_mappings, "includeConsumption": include_consumption,
                  "includeConnectors": include_connectors}

        if sort_by is not None:
            params['sortBy'] = sort_by

        connectors = []
        set_param = True
        count = offset
        total_hits = float("inf")
        if save_results:
            file = open(filename, 'w', encoding='utf8')
        while count < total_hits:
            conn = self.carol.call_api('v1/connectors', params=params)
            count += conn['count']
            if set_param:
                total_hits = conn["totalHits"]
                set_param = False
            conn = conn['hits']

            if not conn:
                break

            connectors.extend(conn)
            params['offset'] = count
            if print_status:
                print('{}/{}'.format(count, total_hits), end='\r')
            if save_results:
                file.write(json.dumps(conn, ensure_ascii=False))
                file.write('\n')
                file.flush()
        if save_results:
            file.close()
        return connectors

    def stats(self, connector_id=None, connector_name=None, all_connectors=False):

        if all_connectors:
            response = self.carol.call_api('v1/connectors/stats/all')
        else:
            if connector_name:
                connector_id = self.get_by_name(connector_name)['mdmId']
            else:
                assert connector_id

            response = self.carol.call_api('v1/connectors/{}/stats'.format(connector_id))

        self._conn_stats = response['aggs']
        return {key: list(value['stagingEntityStats'].keys()) for key, value in self._conn_stats.items()}

    def staging_to_connectors_map(self):
        d = defaultdict(list)
        connectors = self.get_all(print_status=False)
        for connector in connectors:
            current_connector = connector['mdmId']
            conn_stats = self.stats(current_connector)
            for i in conn_stats[current_connector]:
                d[i].append(current_connector)

        return d

    def find_by_staging(self, staging_name=None):
        d = self.staging_to_connectors_map()

        if staging_name:
            conn = d.get(staging_name, None)
            if conn is None:
                raise ValueError('There is no staging named {}'.format(staging_name))

            elif len(conn) > 1:
                print('More than one connector with the staging {}'.format(staging_name))
                return conn

    def get_dm_mappings(self, connector_id=None, connector_name=None, staging_name=None,
                        dm_id=None, dm_name=None, reverse_mapping=False, offset=0, page_size=1000, sort_by=None,
                        sort_order='ASC', all_connectors=False):
        """
        Get data models mappings information.

        :param connector_id: `str`, default `None`
            Connector Id
        :param connector_name: `str`, default `None`
            Connectot name
        :param staging_name: `str`, default `None`
            Staging name
        :param dm_id:  `str`, default `None`
            Data model Id
        :param dm_name: `str`, default `None`
            Data model name
        :param reverse_mapping: `bool`, default `False`
            If to return the reverse mapping.
        :param offset: `int`, default `0`
            Offset for the response.
        :param page_size: `int`, default `1000`
            number of records in each call.
        :param sort_by: `str`, default `None`
            Sort response by
        :param sort_order: `str`, default `ASC`
            Sort Order. Possible values "ASC" and "DESC"
        :param all_connectors: `bool`, default `False`
            It will return all the mapping for all connectors/stagings
        :return:
        """

        if all_connectors:
            payload = {
                "offset": offset,
                "sortBy": sort_by,
                "pageSize": page_size,
                "sortOrder": sort_order
            }
            url = "v1/connectors/mappings/all"

        else:
            if connector_name:
                connector_id = self.get_by_name(connector_name)['mdmId']
            else:
                assert connector_id

            if dm_name is not None:
                url_dm = f"v1/entities/templates/name/{dm_name}"
                dm_id = self.carol.call_api(url_dm, method='GET')['mdmId']

            payload = {
                "reverseMapping": reverse_mapping,
                "entityId": dm_id,
                "stagingType": staging_name,
                "offset": offset,
                "sortBy": sort_by,
                "pageSize": page_size,
                "sortOrder": sort_order
            }

            url = f"v1/connectors/{connector_id}/entityMappings"
        set_param = True
        to_get = float('inf')
        count = 0
        self.resp = []
        while count < to_get:

            resp = self.carol.call_api(url, method='GET', params=payload)
            if set_param:
                self.total_hits = resp["totalHits"]
                to_get = resp["totalHits"]
                set_param = False

            count += resp['count']
            query = resp['hits']
            if not query:
                break

            self.resp.extend(query)
            payload['offset'] = count

        return self.resp



import bokeh
assert bokeh.__version__ == '1.2.0'
from bokeh.models import (
    ColumnDataSource,
    BooleanFilter,
    HoverTool,
)

from bokeh.events import ButtonClick
from bokeh.models import Button
from bokeh.layouts import widgetbox
from bokeh.models.widgets import PreText

from bokeh.layouts import column, row, layout
from bokeh.plotting import figure
from bokeh.transform import transform
from bokeh.palettes import Category10
from bokeh.models.mappers import CategoricalColorMapper

# Annotation imports
from typing import Type, List
from pycarol.pipeline import Pipe, Task


def _make_colormapper(data_source: dict, col_name: str):
    import numpy as np
    factors = list(set(data_source[col_name]))
    nb_factors = np.clip(len(factors), 3, 10)
    return CategoricalColorMapper(factors=factors,
                                  palette=Category10[nb_factors])



def _make_pipeline_plot(
        nodes_data_source: Type[ColumnDataSource],
        edges_data_source: Type[ColumnDataSource],
):
    def _update_incomplete(nodes_data_source,incomplete_nodes_data_source):
        nodes_dict = nodes_data_source.data
        incomplete_mask = [(not c) for c in nodes_dict['complete']]
        incomplete_dict = {
            'x': [x for i,x in enumerate(nodes_dict['x']) if incomplete_mask[i]],
            'y': [x for i,x in enumerate(nodes_dict['y']) if incomplete_mask[i]]
            }
        incomplete_nodes_data_source.data = incomplete_dict

    incomplete_nodes_data_source = ColumnDataSource()
    _update_incomplete(nodes_data_source,incomplete_nodes_data_source)
    nodes_data_source.on_change(
        'data',
        lambda attr,old,new: _update_incomplete(nodes_data_source,incomplete_nodes_data_source)
        )

    family_color = _make_colormapper(nodes_data_source.data,'task_family')
    fig = figure(
        title="Pipeline Debugger",
        x_range=(-1, max(nodes_data_source.data['x']) + 1),
        y_range=(-1, max(nodes_data_source.data['y']) + 1),
        tools=[
            HoverTool(names=['alltasks']),
            'wheel_zoom',
            'pan',
            'tap',
            'box_select',
            'reset'
        ],
        toolbar_location='right',
        tooltips=[
            ('index', "$index"),
            ('task_id', '@task_id'),
            ('task_family','@task_family'),
            ('complete','@complete'),
            ('target_version', '@target_version'),
            ('task_version', '@task_version'), ],
        plot_width=900,
        plot_height=500,
    )
    fig.xgrid.visible = False
    fig.ygrid.visible = False
    fig.xaxis.visible = False
    fig.yaxis.visible = False
    fig.background_fill_color = "#D3D3D3"

    edges_glyph = fig.segment(
        'x0',
        'y0',
        'x1',
        'y1',
        source=edges_data_source,
        color='gray',
        line_alpha=0.3,
    )
    
    fig.circle(
        'x',
        'y',
        source=nodes_data_source,
        size=20,
        name='alltasks',
        color=transform('task_family', family_color),
        legend='task_family',
    )

    fig.circle(
        'x',
        'y',
        source=incomplete_nodes_data_source,
        size=10,
        color='white',
    )
    
    fig.text(
        x='x',
        y='y',
        text='task_name',
        source=nodes_data_source,
        text_font_size='8pt',
    )

    edges_glyph.nonselection_glyph = None
    return fig

#TODO: WIP. make plotsdynamics abstract class with button decorator and so on
class PlotDynamics():
    def __init__(
            self,
            nodes_data_source: Type[ColumnDataSource],
            edges_data_source: Type[ColumnDataSource],
            pipe: Type[Pipe],
    ):
        #  Note that, although we consider datasources to be instance attributes, a ColumnDataSource can always be externally modified
        self.nodes_data_source = nodes_data_source
        self.edges_data_source = edges_data_source
        assert isinstance(pipe, Pipe)
        self.pipe = pipe
        self.selected_nodes = []
        
        # === Dynamics ===
        self.nodes_data_source.selected.on_change('indices', self.select_callback)

        self.buttons = []
        self.cb_fn_and_names = [
            (self.remove_callback, "Remove Selected"),
            (self.removeupstream_callback, "Remove Upstream"),
            (self.run_callback, "Run"),
            (self.update_callback, "Update"),
            (self.close_callback, "Close"),
        ]
        for cb_fn, name in self.cb_fn_and_names:
            button = Button(label = name)
            button.on_event(ButtonClick,cb_fn)
            self.buttons.append(button)


    def get_selected_tasks(self) -> List[Type[Task]]:
        task_id_column = self.nodes_data_source.data['task_id']
        task_ids = [task_id_column[i] for i in self.selected_nodes]
        selected_tasks = [self.pipe.get_task_by_id(id) for id in task_ids]
        return selected_tasks

    def select_callback(self,attr, old, new):
        self.selected_nodes = new
        print("selected {}".format(new))


    def run_callback(self,event):
        print("entered run callback")
        for t in self.get_selected_tasks():
            print(f"Running {t}")
            t.run()

    def remove_callback(self,event):
        for t in self.get_selected_tasks():
            print(f"Removing {t}")
            t.remove()

    def close_callback(self,event):
        from tornado.ioloop import IOLoop
        ioloop = IOLoop.current()
        ioloop.stop()

    def removeupstream_callback(self,event):
        for t in self.get_selected_tasks():
            assert isinstance(t,Task), f"{t} should be instance of {Task}"
            print(f"Removing upstream {t}")
            self.pipe.remove_upstream([t])
        return

    def update_callback(self,event):
        print("start update")
        self.pipe.update_all_complete_status()
        task_id_column = self.nodes_data_source.data['task_id']
        task_gen = (self.pipe.get_task_by_id(task_id) for task_id in task_id_column)
        update_data = dict(complete=[])
        for task in task_gen:
            update_data['complete'].append(self.pipe.get_task_complete(task))
        print("finish update")

        self.nodes_data_source.data.update(update_data)
        
    def buttons_layout(self):
        return row(self.buttons)

def plot_pipeline(
    nodes_data:dict,
    edges_data:dict,
    pipe: Type[Pipe],
    ):
    """
    Receives data sources in dict format and returns dynamic plot
    Args:
        nodes_data:
        edges_data:

    Returns:
        final_layout:

    """
    #TODO: get nodes_data and edges_data from pipe
    nodes_data_source = ColumnDataSource(data=nodes_data)
    edges_data_source = ColumnDataSource(data=edges_data)


    pipeline_plot = _make_pipeline_plot(nodes_data_source, edges_data_source)

    dynamics = PlotDynamics(nodes_data_source,edges_data_source,pipe)
    dynamics.update_callback(None)
    ### Layout
    final_layout = layout(
        column(
            [
                pipeline_plot,
                row(dynamics.buttons_layout()),
            ],
            sizing_mode='scale_width'
        ),
    )
    return final_layout


def get_plot_from_pipeline(pipe: Type[Pipe]):
    """
    Main module method. From a luigi task with defined parameters, generates a
    bokeh plot of the
    pipeline. It does not render the plot.
    Args:
        pipe: list of luigi task initialised with proper parameters

    Returns:
        bokeh_layout:

    """

    from .viewer import (nodes_layout, edges_layout,
                         make_nodes_data_source, make_edges_data_source,
                         )

    dag = pipe.get_dag()

    nodes_layout = nodes_layout(dag)
    edges_layout = edges_layout(dag, nodes_layout)

    nodes_data_source = make_nodes_data_source(nodes_layout)
    edges_data_source = make_edges_data_source(edges_layout)

    bokeh_layout = plot_pipeline(nodes_data_source, edges_data_source, pipe)

    return bokeh_layout






import dash
import dash_core_components as dcc
import dash_html_components as html
import dash_cytoscape as cyto
cyto.load_extra_layouts()
external_stylesheets = ['https://codepen.io/chriddyp/pen/bWLwgP.css']

def get_app_from_pipeline(pipe):
    from pycarol.pipeline.viewer.viewer import (nodes_layout, edges_layout, )

    app = dash.Dash(__name__, external_stylesheets=external_stylesheets)
    root_tasks = pipe.top_nodes
    dag = pipe.get_dag()

    nodes_elements = [
        {
            'data': {
                'id': k.task_id,
                'label': k.get_task_family().split('.')[-1],
            },
            'classes': k.get_task_family().split('.')[-2],
        }
        for i,k in enumerate(dag.keys())
    ]

    edges_elements =[
        {
            'data': {
                'source': source.task_id,
                'target': target.task_id,
            }
        }
        for i,source in enumerate(dag.keys()) for target in dag[source]
    ]

    group_selectors = [
        {
            'selector': 'node',
            'style': {
                'font-size': '12px',
                'content': 'data(label)',
                'shape': 'oval',
            }
        },
        {
            'selector': 'edge',
            'style': {
                'curve-style': 'bezier',
                'source-arrow-shape': 'vee',
                'arrow-scale': 2,
                # 'line-color': 'blue'
            }
        }
    ]
    import colorcet as cc

    class_selectors = [
        {
            'selector': '.' + family,
            'style': {
                'background-color': cc.glasbey_dark[i],
            }
        }
        for i,family in enumerate(
            {t.get_task_family().split('.')[-2] for t in dag.keys()}
        )
    ]
    my_stylesheet = group_selectors + class_selectors




    app.layout = html.Div(children=[

        html.H1(children='Pipeline viewer'),

        cyto.Cytoscape(
            id='main-pipeline',
            layout=dict(
                name='dagre',
                # rankDir='RL',
                # ranker='longest-path',
                padding=100,
                nodeDimensionsIncludeLabels=True,
                animate=True,
                animationDuration=2000,
            ),
            style=dict(width='100%', height='600px'),
            stylesheet=my_stylesheet,
            elements=nodes_elements+edges_elements,

        )

    ])

    return app



import asyncio
from concurrent.futures import ThreadPoolExecutor

from .miscellaneous import stream_data


def send_a(carol, session, url, data_json, extra_headers, content_type):
    """
    Helper function to be used when sending data async.


    :param carol: requests.Session
        Carol object
    :param session: `requests.Session`
        Session object to handle multiple API calls.
    :param url: `str`
        end point to be called.
    :param data_json: `dict`
        The json to be send.
    :param extra_headers: `dict`
        Extra headers to be used in the API call
    :param content_type: `dict`
        Content type of the call.
    :return: None
    """
    carol.call_api(url, data=data_json, extra_headers=extra_headers,
                   content_type=content_type, session=session)


async def send_data_asynchronous(carol, data, step_size, url, extra_headers,
                                 content_type, max_workers, compress_gzip):
    """

    :param carol: `pycarol.carol.Carol`.
        Carol object
    :param data: `pandas.DataFrame` or `list of dict`,
        Data to be sent.
    :param step_size: 'int'
        Number of records per slice.
    :param url: 'str'
        API URI
    :param extra_headers: `dict`
        Extra headers to be used in the API call
    :param content_type:  `dict`
        Content type of the call.
    :param max_workers:  `int`
        Max number of workers of the async job
    :param compress_gzip: 'bool'
        If to compress the data to send
    :return:
    """
    # based on https://hackernoon.com/how-to-run-asynchronous-web-requests-in-parallel-with-python-3-5-without-aiohttp-264dc0f8546
    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        session = carol._retry_session(status_forcelist=[502, 429, 524, 408, 504, 598, 520, 503, 500],
                                       method_whitelist=frozenset(['POST']))
        # Set any session parameters here before calling `send_a`
        loop = asyncio.get_event_loop()
        tasks = [
            loop.run_in_executor(
                executor,
                send_a,
                *(carol, session, url, data_json, extra_headers, content_type)
                # Allows us to pass in multiple arguments to `send_a`
            )
            for data_json, _ in stream_data(data=data,
                                            step_size=step_size,
                                            compress_gzip=compress_gzip)
        ]
        for _ in await asyncio.gather(*tasks):
            pass



from ..data_models.data_model_types import DataModelTypeIds
from ..verticals import Verticals
from ..data_models.data_models import CreateDataModel

import random


class DataModelGenerator(object):

    """
    This class will create a Datamodel from a python dictionary. Today, it is not allowed nested fileds in the dictionary.

    Ex:
    ```python
    from pycarol.auth.PwdAuth import PwdAuth
    from pycarol.auth.ApiKeyAuth import ApiKeyAuth
    from pycarol.carol import Carol

    from pycarol.utils.data_model_generator import DataModelGenerator

    json_sample = {'atende_sus': 'Sim',
                   'estabelecimento': 'Hospital Raimundo Chaar',
                   'existente': '42',
                   'tipo_do_estabelecimento': '05 - Hospital Geral',
                   'tipo_do_estabelecimento_resumo': 'Hospitais',
                   'uf': 'AC'}

    login = Carol('foo_tenant, 'foo', auth=PwdAuth('foo@totvs.com.br', 'foo123'))
    la = DataModelGenerator(login)
    la.start(json_sample, dm_name='hfair', profile_title='estabelecimento', publish=True, overwrite=True)
    ```

    """


    def __init__(self, carol):
        """

        :param carol: Carol object
            Carol object.
        """
        self.carol = carol

    def start(self, json_sample, dm_name, publish=True, profile_title=None, overwrite=True, vertical_ids=None,
              vertical_names='retail', entity_template_type_ids=None,
              entity_template_type_names='product', dm_label=None,
              group_name='Others', transaction_dm=False):

        """

        :param json_sample:
        :param dm_name:
        :param publish:
        :param profile_title:
        :param overwrite:
        :param vertical_ids:
        :param vertical_names:
        :param entity_template_type_ids:
        :param entity_template_type_names:
        :param dm_label:
        :param group_name:
        :param transaction_dm:
        :return:
        """

        if publish:
            assert profile_title in json_sample

        if ((vertical_names is None) and (vertical_ids is None)):
            self.verticals_dict = Verticals(self.carol).all()
            vertical_names = random.choice(self.verticals_dict.keys())
            vertical_ids = self.verticals_dict.get(vertical_names)
        if ((entity_template_type_ids is None) and (entity_template_type_names is None)):
            self.entityTemplateTypesDict = DataModelTypeIds(self.carol).all()
            entity_template_type_names = random.choice(self.entityTemplateTypesDict.keys())
            entity_template_type_ids = self.entityTemplateTypesDict.get(entity_template_type_names)

        tDM = CreateDataModel(self.carol)
        tDM.create(dm_name, overwrite=overwrite, vertical_ids=vertical_ids, vertical_names=vertical_names,
                   entity_template_type_ids=entity_template_type_ids,
                   entity_template_type_names=entity_template_type_names,
                   label=dm_label, group_name=group_name, transaction_data_model=transaction_dm)

        dm_id = tDM.template_dict[dm_name]['mdmId']

        tDM.from_json(json_sample, profile_title=profile_title, publish=publish, dm_id=dm_id)
        print('Done!')



# bokeh serve --show web_app.py
import dotenv
dotenv.load_dotenv()

from pycarol.pipeline.tools.pipeline_example import pipeline1

from bokeh.plotting import curdoc
doc = curdoc()
from pycarol.pipeline.viewer.bokeh_plot import get_plot_from_pipeline
plot = get_plot_from_pipeline(pipeline1)
doc.add_root(plot)






class Visualization:
    """ Main class to control Tasks visualizations

    This class is intended to provide the base for Carol pipelines visualization.
    It uses bokeh as a backend and provides some common plots that can be used.

    This class can be extended to provide also user specific visualizations.

    E.g.

    MyVisualization(Visualization):
        def __call__(*args, **kwargs):
            do my plot...

        def plot_another_thing(...):
            do other plot...

    and in a pipeline task:

    MyTask(Task):
        visualization_class = MyVisualization

    """

    def __call__(self, *args, **kwargs):
        raise NotImplementedError('No visualization defined for that task.')

    def __init__(self, task=None):
        self._task = task

    def null_data(self, data=None, *args, **kwargs):
        import missingno as msno
        msno.matrix(df=data, *args, **kwargs)
